/* 
 * MultSockChan.c
 *
 *	Multinet driver for channels based on TCP sockets.
 *
 */

#include "multinet_root:[multinet.include.sys]ioctl.h"
#include "multinet_root:[multinet.include.vms]inetiodef.h"
#include "multinet_root:[multinet.include]errno.h"

#include "tclInt.h"		/* Internal definitions for Tcl. */
#include "tclPort.h"		/* Portability features for Tcl. */

#include <stdio.h>
#include <starlet.h>

#define BIT_IS_SET(w,b)	(w & (b))
#define SET_BIT(w,b)	w |= (b)
#define CLEAR_BIT(w,b)	w &= (~(b))

/*
 * This structure describes per-instance state of a tcp based channel.
 */

typedef struct TcpState {
    int flags;				/* ORed combination of the
                                         * bitfields defined below. */
    Tcl_File sock;			/* The socket itself. */
    Tcl_TcpAcceptProc *acceptProc;	/* Proc to call on accept. */
    ClientData acceptProcData;		/* The data for the accept proc. */

    int chan;
    int eventFlag;
    int checkMask;
    int watchMask;
    short iosb[4];
} TcpState;

/*
 * These bits may be ORed together into the "flags" field of a TcpState
 * structure.
 */

#define TCP_ASYNC_SOCKET	(1<<0)
#define TCP_ASYNC_CONNECT	(1<<1)
#define TCP_WATCHING_SOCKET	(1<<2)
#define TCP_CHAN_READY		(1<<3)
#define TCP_SELECT_PENDING	(1<<4)
#define TCP_CHAN_READY		(1<<3)
#define TCP_SELECT_PENDING	(1<<4)

/*
 * Static routines for this file:
 */

static TcpState *	CreateSocket _ANSI_ARGS_((Tcl_Interp *interp,
			    int port, char *host, int server,
			    char *myaddr, int myport, int async));
static int		CreateSocketAddress _ANSI_ARGS_(
			    (struct sockaddr_in *sockaddrPtr,
			    char *host, int port));
static void		TcpAccept _ANSI_ARGS_((ClientData data, int mask));
static int		TcpBlockModeProc _ANSI_ARGS_((ClientData data,
		            Tcl_File inFile, Tcl_File outFile, int mode));
static int		TcpCloseProc _ANSI_ARGS_((ClientData instanceData,
			    Tcl_Interp *interp, Tcl_File inFile,
			    Tcl_File outFile));
static int		TcpInputProc _ANSI_ARGS_((ClientData instanceData,
		            Tcl_File infile, char *buf, int toRead,
        		    int *errorCode));
static int		TcpOutputProc _ANSI_ARGS_((ClientData instanceData,
		            Tcl_File outFile, char *buf, int toWrite,
		            int *errorCode));
static int		TcpGetOptionProc _ANSI_ARGS_((ClientData instanceData,
                            char *optionName, Tcl_DString *dsPtr));

static void		socketReadyHandler(TcpState *);
static void		socketFreeProc(ClientData);

/*
 * This structure describes the channel type structure for TCP socket
 * based IO:
 */

static Tcl_ChannelType tcpChannelType = {
    "tcp",				/* Type name. */
    TcpBlockModeProc,			/* Set blocking/nonblocking mode.*/
    TcpCloseProc,			/* Close proc. */
    TcpInputProc,			/* Input proc. */
    TcpOutputProc,			/* Output proc. */
    NULL,				/* Seek proc. */
    NULL,				/* Set option proc. */
    TcpGetOptionProc			/* Get option proc. */
};

/*
 * The preferred buffer size for Unix channels.
 */

#define SOCKET_BUFSIZE	4096


/*
 *----------------------------------------------------------------------
 *
 * TcpBlockModeProc --
 *
 *	Set blocking or non-blocking mode on a TCP channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpBlockModeProc(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Socket state. */
    Tcl_File inFile, outFile;		/* Input, output for channel. */
    int mode;				/* The mode to set. */
{
    unsigned int enable;
    int sock;
    TcpState *statePtr;
    
    statePtr = (TcpState *) instanceData;
    if (mode == TCL_MODE_BLOCKING) {
        CLEAR_BIT(statePtr->flags, TCP_ASYNC_SOCKET);
    }
    else {
        SET_BIT(statePtr->flags, TCP_ASYNC_SOCKET);
    }
    enable = (mode == TCL_MODE_BLOCKING) ? 0 : 1;
    if (inFile != NULL) {
        sock = (int) Tcl_GetFileInfo(inFile, NULL);
	if (socket_ioctl(sock, FIONBIO, &enable) < 0)
	    return socket_errno;
    }
    if (outFile != NULL) {
        sock = (int) Tcl_GetFileInfo(outFile, NULL);
	if (socket_ioctl(sock, FIONBIO, &enable) < 0)
	    return socket_errno;
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * WaitForConnect --
 *
 *	Waits for a connection on an asynchronously opened socket to
 *	be completed.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The socket is connected after this function returns.
 *
 *----------------------------------------------------------------------
 */

static int
WaitForConnect(statePtr, fileToWaitFor, errorCodePtr)
    TcpState *statePtr;		/* State of the socket. */
    Tcl_File fileToWaitFor;	/* File to wait on to become connected. */
    int *errorCodePtr;		/* Where to store errors? */
{
    int sock;			/* The socket itself */
    int timeOut;		/* How long to wait. */
    int state;			/* Of calling TclWaitForFile. */
    int enable;

    /*
     * If an asynchronous connect is in progress, attempt to wait for it
     * to complete before reading.
     */
    
    if (BIT_IS_SET(statePtr->flags, TCP_ASYNC_CONNECT)) {
        if (BIT_IS_SET(statePtr->flags, TCP_ASYNC_SOCKET)) {
            timeOut = 0;
        }
        else {
            timeOut = -1;
        }
	socket_errno = 0 ;
        state = TclWaitForFile(fileToWaitFor, TCL_WRITABLE | TCL_EXCEPTION,
		timeOut);
	if (!BIT_IS_SET(statePtr->flags, TCP_ASYNC_SOCKET)) {
	    sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
	    enable = 1;
	    socket_ioctl(sock, FIONBIO, &enable);
	}
	if (BIT_IS_SET(state, TCL_EXCEPTION)) {
	    return -1;
	}
        if (BIT_IS_SET(state, TCL_WRITABLE)) {
            CLEAR_BIT(statePtr->flags, TCP_ASYNC_CONNECT);
        }
         else if (timeOut == 0) {
            *errorCodePtr = errno = socket_errno = EWOULDBLOCK;
            return -1;
        }
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpInputProc --
 *
 *	This procedure is invoked by the generic IO level to read input
 *	from a TCP socket based channel.
 *
 *	NOTE: We cannot share code with FilePipeInputProc because here
 *	we must use recv to obtain the input from the channel, not read.
 *
 * Results:
 *	The number of bytes read is returned or -1 on error. An output
 *	argument contains the POSIX error code on error, or zero if no
 *	error occurred.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpInputProc(instanceData, inFile, buf, bufSize, errorCodePtr)
    ClientData instanceData;		/* Socket state. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCodePtr;			/* Where to store error code. */
{
    TcpState *statePtr;			/* The state of the socket. */
    int sock;				/* The OS handle. */
    int bytesRead;			/* How many bytes were read? */
    int state;				/* Of waiting for connection. */

    *errorCodePtr = 0;
    sock = (int) Tcl_GetFileInfo(inFile, NULL);
    statePtr = (TcpState *) instanceData;

    state = WaitForConnect(statePtr, inFile, errorCodePtr);
    if (state != 0) {
    	CLEAR_BIT(statePtr->flags, TCP_CHAN_READY);
	return -1;
    }
    bytesRead = multinet_recv(sock, buf, bufSize, 0);
    if (bytesRead > -1) {
	CLEAR_BIT(statePtr->flags, TCP_CHAN_READY);
	return bytesRead;
    }
    if (socket_errno == ECONNRESET) {

        /*
         * Turn ECONNRESET into a soft EOF condition.
         */
        
        return 0;
    }
    *errorCodePtr = socket_errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpOutputProc--
 *
 *	This procedure is invoked by the generic IO level to write output
 *	to a TCP socket based channel.
 *
 *	NOTE: We cannot share code with FilePipeOutputProc because here
 *	we must use send, not write, to get reliable error reporting.
 *
 * Results:
 *	The number of bytes written is returned. An output argument is
 *	set to a POSIX error code if an error occurred, or zero.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

static int
TcpOutputProc(instanceData, outFile, buf, toWrite, errorCodePtr)
    ClientData instanceData;		/* Socket state. */
    Tcl_File outFile;			/* Output device for channel. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCodePtr;			/* Where to store error code. */
{
    TcpState *statePtr;
    int written;
    int sock;
    int state;				/* Of waiting for connection. */

    *errorCodePtr = 0;
    sock = (int) Tcl_GetFileInfo(outFile, NULL);
    statePtr = (TcpState *) instanceData;
    state = WaitForConnect(statePtr, outFile, errorCodePtr);
    if (state != 0) {
	return -1;
    }
    written = socket_write(sock, buf, toWrite);
    if (written > -1) {
        return written;
    }
    *errorCodePtr = socket_errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpCloseProc --
 *
 *	This procedure is invoked by the generic IO level to perform
 *	channel-type-specific cleanup when a TCP socket based channel
 *	is closed.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the socket of the channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpCloseProc(instanceData, interp, inFile, outFile)
    ClientData instanceData;	/* The socket to close. */
    Tcl_Interp *interp;		/* For error reporting - unused. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    TcpState *statePtr;
    Tcl_File sockFile;
    int sock;
    int errorCode = 0;

    statePtr = (TcpState *) instanceData;
    sockFile = statePtr->sock;
    sock = (int) Tcl_GetFileInfo(sockFile, NULL);
    
    /*
     * Delete a file handler that may be active for this socket if this
     * is a server socket - the file handler was created automatically
     * by Tcl as part of the mechanism to accept new client connections.
     * Channel handlers are already deleted in the generic IO channel
     * closing code that called this function, so we do not have to
     * delete them here.
     */
    
    Tcl_DeleteFileHandler(sockFile);

    /* freeProc should do it */
    /* ckfree((char *) statePtr); */

    /*
     * We assume that inFile==outFile==sockFile and so
     * we only clean up sockFile.
     */

    Tcl_FreeFile(sockFile);

    if (socket_close(sock) < 0) {
	errorCode = socket_errno;
    }

    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpGetOptionProc --
 *
 *	Computes an option value for a TCP socket based channel, or a
 *	list of all options and their values.
 *
 *	Note: This code is based on code contributed by John Haxby.
 *
 * Results:
 *	A standard Tcl result. The value of the specified option or a
 *	list of all options and	their values is returned in the
 *	supplied DString.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
TcpGetOptionProc(instanceData, optionName, dsPtr)
    ClientData instanceData;		/* Socket state. */
    char *optionName;			/* Name of the option to
                                         * retrieve the value for, or
                                         * NULL to get all options and
                                         * their values. */
    Tcl_DString *dsPtr;			/* Where to store the computed
                                         * value; initialized by caller. */
{
    TcpState *statePtr;
    struct sockaddr_in sockname;
    struct sockaddr_in peername;
    struct hostent *hostEntPtr;
    int sock;
    int size = sizeof(struct sockaddr_in);
    size_t len = 0;
    char buf[128];

    statePtr = (TcpState *) instanceData;
    sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
    if (optionName != (char *) NULL) {
        len = strlen(optionName);
    }

    if ((len == 0) ||
            ((len > 1) && (optionName[1] == 'p') &&
                    (strncmp(optionName, "-peername", len) == 0))) {
        if (multinet_getpeername(sock, (struct sockaddr *) &peername, &size) >= 0) {
            if (len == 0) {
                Tcl_DStringAppendElement(dsPtr, "-peername");
                Tcl_DStringStartSublist(dsPtr);
            }
            Tcl_DStringAppendElement(dsPtr, multinet_inet_ntoa(peername.sin_addr));
            hostEntPtr = multinet_gethostbyaddr((char *) &(peername.sin_addr),
                    sizeof(peername.sin_addr), AF_INET);
            if (hostEntPtr != (struct hostent *) NULL) {
                Tcl_DStringAppendElement(dsPtr, hostEntPtr->h_name);
            } else {
                Tcl_DStringAppendElement(dsPtr, multinet_inet_ntoa(peername.sin_addr));
            }
            sprintf(buf, "%d", ntohs(peername.sin_port));
            Tcl_DStringAppendElement(dsPtr, buf);
            if (len == 0) {
                Tcl_DStringEndSublist(dsPtr);
            } else {
                return TCL_OK;
            }
        }
    }

    if ((len == 0) ||
            ((len > 1) && (optionName[1] == 's') &&
                    (strncmp(optionName, "-sockname", len) == 0))) {
        if (multinet_getsockname(sock, (struct sockaddr *) &sockname, &size) >= 0) {
            if (len == 0) {
                Tcl_DStringAppendElement(dsPtr, "-sockname");
                Tcl_DStringStartSublist(dsPtr);
            }
            Tcl_DStringAppendElement(dsPtr, multinet_inet_ntoa(sockname.sin_addr));
            hostEntPtr = multinet_gethostbyaddr((char *) &(sockname.sin_addr),
                    sizeof(peername.sin_addr), AF_INET);
            if (hostEntPtr != (struct hostent *) NULL) {
                Tcl_DStringAppendElement(dsPtr, hostEntPtr->h_name);
            } else {
                Tcl_DStringAppendElement(dsPtr, multinet_inet_ntoa(sockname.sin_addr));
            }
            sprintf(buf, "%d", ntohs(sockname.sin_port));
            Tcl_DStringAppendElement(dsPtr, buf);
            if (len == 0) {
                Tcl_DStringEndSublist(dsPtr);
            } else {
                return TCL_OK;
            }
        }
    }

    if (len > 0) {
	socket_errno = EINVAL;
        Tcl_SetErrno(EINVAL);
        return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateSocket --
 *
 *	This function opens a new socket in client or server mode
 *	and initializes the TcpState structure.
 *
 * Results:
 *	Returns a new TcpState, or NULL with an error in interp->result,
 *	if interp is not NULL.
 *
 * Side effects:
 *	Opens a socket.
 *
 *----------------------------------------------------------------------
 */

static TcpState *
CreateSocket(interp, port, host, server, myaddr, myport, async)
    Tcl_Interp *interp;		/* For error reporting; can be NULL. */
    int port;			/* Port number to open. */
    char *host;			/* Name of host on which to open port.
				 *NULL implies INADDR_ANY */
    int server;			/* 1 if socket should be a server socket,
				 * else 0 for a client socket. */
    char *myaddr;		/* Optional client-side address */
    int myport;			/* Optional client-side port */
    int async;			/* If nonzero and creating a client socket,
				 * attempt to do an async connect. Otherwise
				 * do a synchronous connect or bind. */
{
    char *msg = "couldn't open socket: ";
    unsigned int enable;
    int status, sock, asyncConnect, curState, origState;
    struct sockaddr_in sockaddr;	/* Socket address */
    struct sockaddr_in mysockaddr;	/* Socket address for client */
    TcpState *statePtr;

    sock = -1;
    origState = 0;
    if (! CreateSocketAddress(&sockaddr, host, port)) {
    	goto addressError;
    }
    if ((myaddr != NULL || myport != 0) &&
    	    ! CreateSocketAddress(&mysockaddr, myaddr, myport)) {
	msg = "socket address error";
	goto addressError;
    }

    sock = multinet_socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	goto addressError;
    }

    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, SOCKET_BUFSIZE);

    asyncConnect = 0;
    status = 0;
    if (server) {

	/*
	 * Set up to reuse server addresses automatically and bind to the
	 * specified port.
	 */
    
	status = 1;
	(void) multinet_setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
		sizeof(status));
	status = multinet_bind(sock, (struct sockaddr *) &sockaddr,
                sizeof(struct sockaddr));
	if (status != -1) {
	    status = multinet_listen(sock, 5);
	} 
    } else {
    	if (myaddr != NULL || myport != 0) {
    	    status = 1;
    	    (void) multinet_setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
    	    	sizeof(status));
	    status = multinet_bind(sock, (struct sockaddr *) &mysockaddr,
	    	    sizeof(struct sockaddr));
	    if (status < 0) {
	    	msg = "couldn't bind socket: ";
	    	goto bindError;
	    }
	}

	/*
	 * Attempt to connect. The connect may fail at present with an
	 * EINPROGRESS but at a later time it will complete. The caller
	 * will set up a file handler on the socket if she is interested in
	 * being informed when the connect completes.
	 */

	if (async) {
	    enable = 1;
	    status = socket_ioctl(sock, FIONBIO, &enable);
	}
	else {
	    status = 0;
	}
        if (status > -1) {
            status = multinet_connect(sock, (struct sockaddr *) &sockaddr,
                    sizeof(sockaddr));
            if (status < 0) {
                if (socket_errno == EINPROGRESS) {
                    asyncConnect = 1;
                    status = 0;
                }
            }
            if ((async) && status == 0) {
		enable = 0;
		status = socket_ioctl(sock, FIONBIO, &enable);
            }
        }
        else
            msg = "couldn't set non-blocking mode on socket: ";
    }

bindError:
    if (status < 0) {
        if (interp != NULL) {
            Tcl_AppendResult(interp, msg,
                    vms_errno_string(), (char *) NULL);
        }
	if (sock != -1) {
	    socket_close(sock);
	}
        return NULL;
    }

    /*
     * Allocate a new TcpState for this socket.
     */

    statePtr = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    statePtr->flags = 0;
    if (asyncConnect) {
        SET_BIT(statePtr->flags, TCP_ASYNC_CONNECT);
    }
    statePtr->sock = Tcl_GetFile((ClientData) sock, TCL_VMS_SOCKET);
    statePtr->chan = sock;
    statePtr->eventFlag = 0;
    statePtr->checkMask = 0;
    statePtr->watchMask = 0;
    Tcl_SetNotifierData(statePtr->sock, socketFreeProc, (ClientData)statePtr);
    
    return statePtr;

addressError:
    if (sock != -1) {
	socket_close(sock);
    }
    if (interp != NULL) {
	Tcl_AppendResult(interp, msg,
		vms_errno_string(), (char *) NULL);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateSocketAddress --
 *
 *	This function initializes a sockaddr structure for a host and port.
 *
 * Results:
 *	1 if the host was valid, 0 if the host could not be converted to
 *	an IP address.
 *
 * Side effects:
 *	Fills in the *sockaddrPtr structure.
 *
 *----------------------------------------------------------------------
 */

static int
CreateSocketAddress(sockaddrPtr, host, port)
    struct sockaddr_in *sockaddrPtr;	/* Socket address */
    char *host;				/* Host.  NULL implies INADDR_ANY */
    int port;				/* Port number */
{
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */

    (void) memset((char *) sockaddrPtr, '\0', sizeof(struct sockaddr_in));
    sockaddrPtr->sin_family = AF_INET;
    sockaddrPtr->sin_port = htons((unsigned short) (port & 0xFFFF));
    if (host == NULL) {
	addr.s_addr = INADDR_ANY;
    } else {
	addr.s_addr = multinet_inet_addr(host);
	if (addr.s_addr == (unsigned long) -1) {
	    hostent = multinet_gethostbyname(host);
	    if (hostent != NULL) {
		memcpy((void *) &addr,
			(void *) hostent->h_addr_list[0],
			(size_t) hostent->h_length);
	    } else {
#ifdef	EHOSTUNREACH
		socket_errno = errno = EHOSTUNREACH;
#else
#ifdef	ENXIO
		socket_errno = errno = ENXIO;
#endif
#endif
		return 0;	/* error */
	    }
	}
    }

    /*
     * NOTE: On 64 bit machines the assignment below is rumored to not
     * do the right thing. Please report errors related to this if you
     * observe incorrect behavior on 64 bit machines such as DEC Alphas.
     * Should we modify this code to do an explicit memcpy?
     */

    sockaddrPtr->sin_addr.s_addr = addr.s_addr;
    return 1;	/* Succcess. */
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpClient --
 *
 *	Opens a TCP client socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.  An error message is returned
 *	in the interpreter on failure.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpClient(interp, port, host, myaddr, myport, async)
    Tcl_Interp *interp;			/* For error reporting; can be NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Host on which to open port. */
    char *myaddr;			/* Client-side address */
    int myport;				/* Client-side port */
    int async;				/* If nonzero, attempt to do an
					 * asynchronous connect. Otherwise
					 * we do a blocking connect. */
{
    Tcl_Channel chan;
    TcpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateSocket(interp, port, host, 0, myaddr, myport, async);
    if (statePtr == NULL) {
	return NULL;
    }

    statePtr->acceptProc = NULL;
    statePtr->acceptProcData = (ClientData) NULL;

    sprintf(channelName, "sock%d",
	    (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, statePtr->sock,
	    statePtr->sock, (ClientData) statePtr);
    if (Tcl_SetChannelOption(interp, chan, "-translation", "auto crlf") ==
	    TCL_ERROR) {
	Tcl_Close((Tcl_Interp *) NULL, chan);
	return NULL;
    }

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_MakeTcpClientChannel --
 *
 *	Creates a Tcl_Channel from an existing client TCP socket.
 *
 * Results:
 *	The Tcl_Channel wrapped around the preexisting TCP socket.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_MakeTcpClientChannel(sock)
    ClientData sock;		/* The socket to wrap up into a channel. */
{
    TcpState *statePtr;
    Tcl_File sockFile;
    char channelName[20];
    Tcl_Channel chan;

    sockFile = Tcl_GetFile(sock, TCL_VMS_SOCKET);
    statePtr = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    statePtr->sock = sockFile;
    statePtr->acceptProc = NULL;
    statePtr->acceptProcData = (ClientData) NULL;
    statePtr->flags = 0;
    statePtr->chan = (int)sock;
    statePtr->eventFlag = 0;
    statePtr->checkMask = 0;
    statePtr->watchMask = 0;
    Tcl_SetNotifierData(statePtr->sock, socketFreeProc, (ClientData)statePtr);

    sprintf(channelName, "sock%d", (int) sock);
    
    chan = Tcl_CreateChannel(&tcpChannelType, channelName, sockFile, sockFile,
            (ClientData) statePtr);
    if (Tcl_SetChannelOption((Tcl_Interp *) NULL, chan, "-translation",
	    "auto crlf") == TCL_ERROR) {
	Tcl_Close((Tcl_Interp *) NULL, chan);
	return NULL;
    }
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpServer --
 *
 *	Opens a TCP server socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed. If an error occurred, an
 *	error message is left in interp->result if interp is
 *	not NULL.
 *
 * Side effects:
 *	Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpServer(interp, port, myHost, acceptProc, acceptProcData)
    Tcl_Interp *interp;			/* For error reporting - may be
                                         * NULL. */
    int port;				/* Port number to open. */
    char *myHost;			/* Name of local host. */
    Tcl_TcpAcceptProc *acceptProc;	/* Callback for accepting connections
                                         * from new clients. */
    ClientData acceptProcData;		/* Data for the callback. */
{
    Tcl_Channel chan;
    TcpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateSocket(interp, port, myHost, 1, NULL, 0, 0);
    if (statePtr == NULL) {
	return NULL;
    }

    statePtr->acceptProc = acceptProc;
    statePtr->acceptProcData = acceptProcData;

    /*
     * Set up the callback mechanism for accepting connections
     * from new clients.
     */

    Tcl_CreateFileHandler(statePtr->sock, TCL_READABLE, TcpAccept,
            (ClientData) statePtr);
    sprintf(channelName, "sock%d",
	    (int) Tcl_GetFileInfo(statePtr->sock, NULL));
    chan = Tcl_CreateChannel(&tcpChannelType, channelName, NULL, NULL,
            (ClientData) statePtr);
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpAccept --
 *	Accept a TCP socket connection.  This is called by the event loop.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Creates a new connection socket. Calls the registered callback
 *	for the connection acceptance mechanism.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
TcpAccept(data, mask)
    ClientData data;			/* Callback token. */
    int mask;				/* Not used. */
{
    TcpState *sockState;		/* Client data of server socket. */
    int newsock;			/* The new client socket */
    Tcl_File newFile;			/* Its file. */
    TcpState *newSockState;		/* State for new socket. */
    struct sockaddr_in addr;		/* The remote address */
    int len;				/* For accept interface */
    Tcl_Channel chan;			/* Channel instance created. */
    char channelName[20];

    sockState = (TcpState *) data;

    len = sizeof(struct sockaddr_in);
    newsock = multinet_accept((int) Tcl_GetFileInfo(sockState->sock, NULL),
	    (struct sockaddr *)&addr, (unsigned int *)&len);
    if (newsock < 0) {
        return;
    }
    
    CLEAR_BIT(sockState->flags, TCP_CHAN_READY);
    newFile = Tcl_GetFile((ClientData) newsock, TCL_VMS_SOCKET);
    if (newFile) {
        newSockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));

        newSockState->flags = 0;
        newSockState->sock = newFile;
        newSockState->acceptProc = (Tcl_TcpAcceptProc *) NULL;
        newSockState->acceptProcData = (ClientData) NULL;
	newSockState->chan = newsock;
	newSockState->eventFlag = 0;
	newSockState->checkMask = 0;
	newSockState->watchMask = 0;
	Tcl_SetNotifierData(newSockState->sock, socketFreeProc, (ClientData)newSockState);

        sprintf(channelName, "sock%d", (int) newsock);
        chan = Tcl_CreateChannel(&tcpChannelType, channelName, newFile,
                newFile, (ClientData) newSockState);
        if (chan == (Tcl_Channel) NULL) {
	    if (newSockState->eventFlag)
		freeEventFlag(newSockState->eventFlag);
            ckfree((char *) newSockState);
            socket_close(newsock);
            Tcl_FreeFile(newFile);
	} else {
            if (Tcl_SetChannelOption((Tcl_Interp *) NULL, chan, "-translation",
		    "auto crlf") == TCL_ERROR) {
		Tcl_Close((Tcl_Interp *) NULL, chan);
	    }
            if (sockState->acceptProc != (Tcl_TcpAcceptProc *) NULL) {
                (sockState->acceptProc) (sockState->acceptProcData, chan,
                        multinet_inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
            }
	}
    }
}

static void
socketFreeProc(ClientData clientdata)
{
    /*
     * Free up socket info structure
     */
    TcpState *statePtr = (TcpState *)clientdata;

    if (statePtr) {
	if (statePtr->eventFlag)
	    freeEventFlag(statePtr->eventFlag);
	ckfree((char *)statePtr);
    }
}

#define	SELECT_DONTWAIT		(1<<0)
#define	SELECT_READABLE		(1<<1)
#define	SELECT_WRITEABLE	(1<<2)
#define	SELECT_EXCEPTION	(1<<3)

TclVmsWatchSocket(Tcl_File file, int mask)
{
    /*
     * Watch for something to happen on a socket.
     *
     * Start a QIO and return the event flag
     * that will be set when something happens.
     */
    int flag, status;
    TcpState *statePtr = (TcpState *)Tcl_GetNotifierData(file, NULL);

    if ((flag = statePtr->eventFlag) == 0) {
	flag = statePtr->eventFlag = getEventFlag();
    }
    if (BIT_IS_SET(statePtr->flags, TCP_SELECT_PENDING))
    	return flag;
    SET_BIT(statePtr->flags, TCP_SELECT_PENDING);
    statePtr->watchMask = 0;
    if (BIT_IS_SET(mask, TCL_READABLE))
	SET_BIT(statePtr->watchMask, SELECT_READABLE);
    if (BIT_IS_SET(mask, TCL_WRITABLE))
	SET_BIT(statePtr->watchMask, SELECT_WRITEABLE);
    if (BIT_IS_SET(mask, TCL_EXCEPTION))
	SET_BIT(statePtr->watchMask, SELECT_EXCEPTION);
    status = sys$qio(flag, statePtr->chan, IO$_SELECT, statePtr->iosb,
		     socketReadyHandler, statePtr,
		     &statePtr->watchMask,
		     0, 0, 0, 0, 0);
    if (status & 1)
	SET_BIT(statePtr->flags, TCP_WATCHING_SOCKET);
    else
	printf("IO$SELECT failed: %x\n", status);
    return flag;
}

VmsSocketReady(Tcl_File file, int mask)
{
    /*
     * Check and see what's ready for a Multinet socket
     */
    int readyMask = 0;
    TcpState *statePtr = (TcpState *)Tcl_GetNotifierData(file, NULL);

    if (!BIT_IS_SET(statePtr->flags, TCP_CHAN_READY))
    	return 0;

    /*
     *Convert Multinet flags to TCL flags
     */
    if (BIT_IS_SET(statePtr->watchMask, SELECT_READABLE))
	SET_BIT(readyMask, TCL_READABLE);
    if (BIT_IS_SET(statePtr->watchMask, SELECT_WRITEABLE))
	SET_BIT(readyMask, TCL_WRITABLE);
    if (BIT_IS_SET(statePtr->watchMask, SELECT_EXCEPTION))
	SET_BIT(readyMask, TCL_EXCEPTION);
    return (readyMask & mask);
}

static void
socketReadyHandler(TcpState *statePtr)
{
    /*
     * AST routine called when something is ready on a socket.
     */
    if (statePtr->iosb[0] & 1) {
	CLEAR_BIT(statePtr->flags, TCP_WATCHING_SOCKET);
	SET_BIT(statePtr->flags, TCP_CHAN_READY);
	CLEAR_BIT(statePtr->flags, TCP_SELECT_PENDING);
    }
    else
	printf("IO$SELECT failed(AST): %x\n", statePtr->iosb[1]);
}
